#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import io
import cv2
import copy
import math
import json
import psutil
import shutil
import numpy as np
import pandas as pd
from PIL import Image

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.monitoring.model_logger import ModelLogger
from {{package_name}}.models_training.object_detectors import utils_faster_rcnn
from {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn import (ModelKerasFasterRcnnObjectDetector,
                                                                                       ModelCheckpointAll,
                                                                                       CustomGeneratorRpn,
                                                                                       CustomGeneratorClassifier)

import tensorflow as tf
from tensorflow.keras.layers import Input


# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)
    tf.keras.backend.clear_session()


# Mocks

def download_url_crash(x, y):
    raise ConnectionError("error")


class MockCustomGeneratorRpn():
    def __init__(self, img_data_list: list, batch_size: int, shuffle: bool, seed: int, model, **kwargs) -> None:
        self.n = len(img_data_list)
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed
        self.img_data_list = img_data_list
        self.model = model
        # Default arguments
        self.horizontal_flip = kwargs.get('horizontal_flip', False)
        self.vertical_flip = kwargs.get('vertical_flip', False)
        self.rot_90 = kwargs.get('rot_90', False)
        self.data_type = kwargs.get('data_type', 'train')
        self.with_img_data = kwargs.get('with_img_data', False)

class MockCustomGeneratorClassifier(MockCustomGeneratorRpn):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Default arguments
        self.shared_model_trainable = kwargs.get('shared_model_trainable', False)

class MockModelKerasFasterRcnnObjectDetector():
    def __init__(self) -> None:
        self.anchor_box_sizes = [64, 128, 256]  # In the paper : [128, 256, 512]
        self.anchor_box_ratios = [[1, 1], [1./math.sqrt(2), 2./math.sqrt(2)], [2./math.sqrt(2), 1./math.sqrt(2)]]  # In the paper : [1, 1], [1, 2], [2, 1]]
        self.nb_anchors = len(self.anchor_box_sizes) * len(self.anchor_box_ratios)
        self.list_anchors = [[anchor_size * anchor_ratio[0], anchor_size * anchor_ratio[1]]
                             for anchor_size in self.anchor_box_sizes for anchor_ratio in self.anchor_box_ratios]
        self.shared_model_subsampling = 16
        self.rpn_min_overlap = 0.2
        self.rpn_max_overlap = 0.8
        self.rpn_regr_scaling = 4
        self.rpn_restrict_num_regions = 100
        self.logger = logging.getLogger(__name__)
        self.preprocess_input = self._get_preprocess_input()
        # Manage RPN model
        class FakeModelRpn():
            def __init__(self) -> None:
                pass
            def set_weights(self, model):
                pass
            def get_weights(self):
                return None
        self.model_rpn = FakeModelRpn()
    def _get_preprocess_input(self):
        def preprocess_input(img):
            return img
        return preprocess_input
    def _generate_images_with_bboxes(self, img_data: dict, horizontal_flip: bool = False,
                                     vertical_flip: bool = False, rot_90: bool = False):
        with open(img_data['file_path'], 'rb') as f:
            img = Image.open(io.BytesIO(f.read()))
            if img.mode != 'RGB':
                img = img.convert('RGB')
            img = np.asarray(img)
        bboxes = copy.deepcopy(img_data.get('bboxes', []))  # Empty if test
        original_height, original_width = img.shape[0], img.shape[1]
        img = self.preprocess_input(img)
        resized_height, resized_width = img.shape[0], img.shape[1]
        for bbox in bboxes:
            bbox['x1'] = bbox['x1'] * (resized_width / original_width)
            bbox['x2'] = bbox['x2'] * (resized_width / original_width)
            bbox['y1'] = bbox['y1'] * (resized_height / original_height)
            bbox['y2'] = bbox['y2'] * (resized_height / original_height)
        prepared_data = {'img': img, 'bboxes': bboxes, 'original_height': original_height,
                         'original_width': original_width, 'resized_height': resized_height,
                         'resized_width': resized_width}
        return prepared_data

def mock_get_rpn_targets(model, img_data_batch: list):
    return 'toto', 'titi'

def mock_clone_model_fn(model, input_tensors=None, clone_function=None):
    return model

# Tests

class ModelKerasFasterRcnnObjectDetectorTests(unittest.TestCase):
    '''Main class to test model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector'''


    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test01_model_keras_faster_rcnn_init(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector.__init__'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Init., test all parameters
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertEqual(model.model_type, 'object_detector')
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.model, None)
        self.assertEqual(model.shared_model, None)
        self.assertEqual(model.model_rpn, None)
        self.assertEqual(model.model_classifier, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 8)
        self.assertEqual(model.epochs_classifier_trainable_true, 8)
        self.assertEqual(model.epochs_rpn_trainable_false, 8)
        self.assertEqual(model.epochs_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8, epochs_rpn_trainable_true=2,
                                                   epochs_classifier_trainable_true=3, epochs_rpn_trainable_false=4,
                                                   epochs_classifier_trainable_false=5)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 2)
        self.assertEqual(model.epochs_classifier_trainable_true, 3)
        self.assertEqual(model.epochs_rpn_trainable_false, 4)
        self.assertEqual(model.epochs_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 8)
        self.assertEqual(model.batch_size_classifier_trainable_true, 8)
        self.assertEqual(model.batch_size_rpn_trainable_false, 8)
        self.assertEqual(model.batch_size_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8, batch_size_rpn_trainable_true=2,
                                                   batch_size_classifier_trainable_true=3, batch_size_rpn_trainable_false=4,
                                                   batch_size_classifier_trainable_false=5)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 2)
        self.assertEqual(model.batch_size_classifier_trainable_true, 3)
        self.assertEqual(model.batch_size_rpn_trainable_false, 4)
        self.assertEqual(model.batch_size_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 8)
        self.assertEqual(model.patience_classifier_trainable_true, 8)
        self.assertEqual(model.patience_rpn_trainable_false, 8)
        self.assertEqual(model.patience_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8, patience_rpn_trainable_true=2,
                                                   patience_classifier_trainable_true=3, patience_rpn_trainable_false=4,
                                                   patience_classifier_trainable_false=5)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 2)
        self.assertEqual(model.patience_classifier_trainable_true, 3)
        self.assertEqual(model.patience_rpn_trainable_false, 4)
        self.assertEqual(model.patience_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_rpn_trainable_true=0.2, lr_classifier_trainable_true=0.3,
                                                   lr_rpn_trainable_false=0.4, lr_classifier_trainable_false=0.5)
        self.assertEqual(model.keras_params['lr_rpn_trainable_true'], 0.2)
        self.assertEqual(model.keras_params['lr_classifier_trainable_true'], 0.3)
        self.assertEqual(model.keras_params['lr_rpn_trainable_false'], 0.4)
        self.assertEqual(model.keras_params['lr_classifier_trainable_false'], 0.5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=64)
        self.assertEqual(model.img_min_side_size, 64)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.8, rpn_max_overlap=0.9)
        self.assertEqual(model.rpn_min_overlap, 0.8)
        self.assertEqual(model.rpn_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=8)
        self.assertEqual(model.rpn_restrict_num_regions, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=8)
        self.assertEqual(model.pool_resize_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=8)
        self.assertEqual(model.nb_rois_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=0.8)
        self.assertEqual(model.roi_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=8)
        self.assertEqual(model.nms_max_boxes, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.8, classifier_max_overlap=0.9)
        self.assertEqual(model.classifier_min_overlap, 0.8)
        self.assertEqual(model.classifier_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=0.8)
        self.assertEqual(model.pred_bbox_proba_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=0.8)
        self.assertEqual(model.pred_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'toto': 'titi'})
        self.assertEqual(model.data_augmentation_params, {'toto': 'titi'})
        remove_dir(model_dir)

        # Check errors
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=31)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.3, rpn_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.3, classifier_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='rgba')
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='grayscale')

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.utils.download_url', side_effect=download_url_crash)
    def test02_model_keras_faster_rcnn_init_offline(self, mock_download_url):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector.__init__
        - No access to a base model
        '''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Clean cache path if exists
        cache_path = os.path.join(utils.get_data_path(), 'transfer_learning_weights')
        if os.path.isdir(cache_path):
            shutil.rmtree(cache_path)

        # Init., test all parameters
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertEqual(model.model_type, 'object_detector')
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.model, None)
        self.assertEqual(model.shared_model, None)
        self.assertEqual(model.model_rpn, None)
        self.assertEqual(model.model_classifier, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 8)
        self.assertEqual(model.epochs_classifier_trainable_true, 8)
        self.assertEqual(model.epochs_rpn_trainable_false, 8)
        self.assertEqual(model.epochs_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, epochs=8, epochs_rpn_trainable_true=2,
                                                   epochs_classifier_trainable_true=3, epochs_rpn_trainable_false=4,
                                                   epochs_classifier_trainable_false=5)
        self.assertEqual(model.epochs, 8)
        self.assertEqual(model.epochs_rpn_trainable_true, 2)
        self.assertEqual(model.epochs_classifier_trainable_true, 3)
        self.assertEqual(model.epochs_rpn_trainable_false, 4)
        self.assertEqual(model.epochs_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 8)
        self.assertEqual(model.batch_size_classifier_trainable_true, 8)
        self.assertEqual(model.batch_size_rpn_trainable_false, 8)
        self.assertEqual(model.batch_size_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, batch_size=8, batch_size_rpn_trainable_true=2,
                                                   batch_size_classifier_trainable_true=3, batch_size_rpn_trainable_false=4,
                                                   batch_size_classifier_trainable_false=5)
        self.assertEqual(model.batch_size, 8)
        self.assertEqual(model.batch_size_rpn_trainable_true, 2)
        self.assertEqual(model.batch_size_classifier_trainable_true, 3)
        self.assertEqual(model.batch_size_rpn_trainable_false, 4)
        self.assertEqual(model.batch_size_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 8)
        self.assertEqual(model.patience_classifier_trainable_true, 8)
        self.assertEqual(model.patience_rpn_trainable_false, 8)
        self.assertEqual(model.patience_classifier_trainable_false, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, patience=8, patience_rpn_trainable_true=2,
                                                   patience_classifier_trainable_true=3, patience_rpn_trainable_false=4,
                                                   patience_classifier_trainable_false=5)
        self.assertEqual(model.patience, 8)
        self.assertEqual(model.patience_rpn_trainable_true, 2)
        self.assertEqual(model.patience_classifier_trainable_true, 3)
        self.assertEqual(model.patience_rpn_trainable_false, 4)
        self.assertEqual(model.patience_classifier_trainable_false, 5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_rpn_trainable_true=0.2, lr_classifier_trainable_true=0.3,
                                                   lr_rpn_trainable_false=0.4, lr_classifier_trainable_false=0.5)
        self.assertEqual(model.keras_params['lr_rpn_trainable_true'], 0.2)
        self.assertEqual(model.keras_params['lr_classifier_trainable_true'], 0.3)
        self.assertEqual(model.keras_params['lr_rpn_trainable_false'], 0.4)
        self.assertEqual(model.keras_params['lr_classifier_trainable_false'], 0.5)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=64)
        self.assertEqual(model.img_min_side_size, 64)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.8, rpn_max_overlap=0.9)
        self.assertEqual(model.rpn_min_overlap, 0.8)
        self.assertEqual(model.rpn_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=8)
        self.assertEqual(model.rpn_restrict_num_regions, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=8)
        self.assertEqual(model.pool_resize_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=8)
        self.assertEqual(model.nb_rois_classifier, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=0.8)
        self.assertEqual(model.roi_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=8)
        self.assertEqual(model.nms_max_boxes, 8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.8, classifier_max_overlap=0.9)
        self.assertEqual(model.classifier_min_overlap, 0.8)
        self.assertEqual(model.classifier_max_overlap, 0.9)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=0.8)
        self.assertEqual(model.pred_bbox_proba_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=0.8)
        self.assertEqual(model.pred_nms_overlap_threshold, 0.8)
        remove_dir(model_dir)

        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'toto': 'titi'})
        self.assertEqual(model.data_augmentation_params, {'toto': 'titi'})
        remove_dir(model_dir)

        # Check errors
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=31)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_min_overlap=0.3, rpn_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, rpn_restrict_num_regions=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pool_resize_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nb_rois_classifier=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, roi_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, nms_max_boxes=0)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_max_overlap=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, classifier_min_overlap=0.3, classifier_max_overlap=0.1)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_bbox_proba_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=-0.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, pred_nms_overlap_threshold=1.01)
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='rgba')
        with self.assertRaises(ValueError):
            ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, color_mode='grayscale')

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test03_model_keras_faster_rcnn_get_model(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_model'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        model._get_model()  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test04_model_keras_faster_rcnn_compile_model_rpn(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._compile_model_rpn'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_rpn_trainable_true=0.1)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        model.shared_model, model.model_rpn, model.model_classifier, model.model = model._get_model()
        self.assertAlmostEqual(0.1, float(model.model_rpn.optimizer.lr.numpy()))
        model._compile_model_rpn(model.model_rpn, 0.1234)
        self.assertAlmostEqual(0.1234, float(model.model_rpn.optimizer.lr.numpy()))

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test05_model_keras_faster_rcnn_compile_model_classifier(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._compile_model_classifier'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, lr_classifier_trainable_true=0.1)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        model.shared_model, model.model_rpn, model.model_classifier, model.model = model._get_model()
        self.assertAlmostEqual(0.1, float(model.model_classifier.optimizer.lr.numpy()))
        model._compile_model_classifier(model.model_classifier, 0.1234)
        self.assertAlmostEqual(0.1234, float(model.model_classifier.optimizer.lr.numpy()))

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test06_model_keras_faster_rcnn_get_shared_model_structure(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_shared_model_structure'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        input_img = Input(shape=(None, None, 3), name='input_img')
        model._get_shared_model_structure(input_img)  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test07_model_keras_faster_rcnn_add_rpn_layers(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._add_rpn_layers'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        input_img = Input(shape=(None, None, 3), name='input_img')
        shared_model_layers = model._get_shared_model_structure(input_img)
        model._add_rpn_layers(shared_model_layers)  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test08_model_keras_faster_rcnn_add_classifier_layers(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._add_classifier_layers'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.list_classes = ['orange', 'banana', 'apple']
        model.dict_classes = {i: col for i, col in enumerate(model.list_classes)}
        model.custom_objects = {**utils_faster_rcnn.get_custom_objects_faster_rcnn(model.nb_anchors, len(model.list_classes)), **model.custom_objects}
        input_img = Input(shape=(None, None, 3), name='input_img')
        input_rois = Input(shape=(None, model.nb_rois_classifier), name='input_rois')
        shared_model_layers = model._get_shared_model_structure(input_img)
        rpn_layers = model._add_rpn_layers(shared_model_layers)
        model._add_classifier_layers(shared_model_layers, input_rois)  # Nothing to test in particular ...

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorClassifier', side_effect=MockCustomGeneratorClassifier)
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorRpn', side_effect=MockCustomGeneratorRpn)
    def test09_model_keras_faster_rcnn_get_generator(self, mock_custom_generator_rpn, mock_custom_generator_classifier):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_generator'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars (no need to put the real data, the generator is mocked)
        df = pd.DataFrame({
            'file_path': ['a', 'c', 'e', 'b'],
            'bboxes': [[[1, 2], [4, 2]], [[1, 5]], [[3, 8]], [[5, 2], [4, 3]]],
        })

        # Nominal case - RPN
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': False, 'vertical_flip': False, 'rot_90': False})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='valid', batch_size=10, generator_type='rpn', with_img_data=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 10)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'valid')
        self.assertEqual(generator.with_img_data, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='test', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df[['file_path']], data_type='test', batch_size=5, generator_type='rpn', with_img_data=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        remove_dir(model_dir)

        # Nominal case - Classifier
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': False, 'vertical_flip': False, 'rot_90': False})
        generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'train')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='valid', batch_size=10, generator_type='classifier', with_img_data=True, shared_model_trainable=False)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 10)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'valid')
        self.assertEqual(generator.with_img_data, True)
        self.assertEqual(generator.shared_model_trainable, False)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df, data_type='test', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(generator.img_data_list[0]['bboxes'], [[1, 2], [4, 2]])
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, data_augmentation_params={'horizontal_flip': True, 'vertical_flip': True, 'rot_90': True})
        generator = model._get_generator(df[['file_path']], data_type='test', batch_size=5, generator_type='classifier', with_img_data=False, shared_model_trainable=True)
        self.assertEqual(generator.n, df.shape[0])
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, None)
        self.assertEqual(len(generator.img_data_list), 4)
        self.assertEqual(model.model_name, generator.model.model_name)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.data_type, 'test')
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.shared_model_trainable, True)
        remove_dir(model_dir)

        # Manage errors
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        with self.assertRaises(ValueError):
            generator = model._get_generator(df, data_type='toto', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='toto')
        with self.assertRaises(ValueError):
            generator = model._get_generator(pd.DataFrame(columns=['toto']), data_type='train', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            generator = model._get_generator(pd.DataFrame(columns=['file_path']), data_type='train', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            generator = model._get_generator(pd.DataFrame(columns=['file_path']), data_type='valid', batch_size=5, generator_type='rpn')
        with self.assertRaises(ValueError):
            model.model_type = 'toto'
            generator = model._get_generator(df, data_type='train', batch_size=5, generator_type='rpn')

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('numpy.random.choice')
    def test10_model_keras_faster_rcnn_generate_images_with_bboxes(self, mock_np_random_choice):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._generate_images_with_bboxes'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        im_path = os.path.join(os.getcwd(), 'test_data', 'apple_36.jpg')
        img_data = {'file_path': im_path, 'bboxes': [{'x1': 10, 'x2': 30, 'y1': 10, 'y2': 50}, {'x1': 5, 'x2': 25, 'y1': 10, 'y2': 45}]}
        img_data_simple = {'file_path': im_path}
        with Image.open(im_path) as im:
            w, h = im.size
        resize_w, resize_h = (250, 300)
        # Mock preprocess
        def fake_preprocess(img):
            img = cv2.resize(img, (resize_w, resize_h), interpolation=cv2.INTER_CUBIC)
            return img

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.preprocess_input = fake_preprocess
        prepared_data = model._generate_images_with_bboxes(img_data, horizontal_flip=False, vertical_flip=False, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(len(prepared_data['bboxes']), len(img_data['bboxes']))
        self.assertNotEqual(prepared_data['bboxes'][0]['x1'], img_data['bboxes'][0]['x1'])
        self.assertNotEqual(prepared_data['bboxes'][0]['y1'], img_data['bboxes'][0]['y1'])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        # Without bbox
        prepared_data = model._generate_images_with_bboxes(img_data_simple, horizontal_flip=False, vertical_flip=False, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(prepared_data['bboxes'], [])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        remove_dir(model_dir)

        # Tests augmentation
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.preprocess_input = fake_preprocess
        prepared_data = model._generate_images_with_bboxes(img_data, horizontal_flip=True, vertical_flip=False, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(len(prepared_data['bboxes']), len(img_data['bboxes']))
        self.assertNotEqual(prepared_data['bboxes'][0]['x1'], img_data['bboxes'][0]['x1'])
        self.assertNotEqual(prepared_data['bboxes'][0]['y1'], img_data['bboxes'][0]['y1'])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        #
        prepared_data = model._generate_images_with_bboxes(img_data_simple, horizontal_flip=False, vertical_flip=True, rot_90=False)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(prepared_data['bboxes'], [])
        self.assertEqual(prepared_data['original_height'], h)
        self.assertEqual(prepared_data['original_width'], w)
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        remove_dir(model_dir)
        #
        mock_np_random_choice.return_value = [90]
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        model.preprocess_input = fake_preprocess
        prepared_data = model._generate_images_with_bboxes(img_data, horizontal_flip=False, vertical_flip=False, rot_90=True)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(len(prepared_data['bboxes']), len(img_data['bboxes']))
        self.assertNotEqual(prepared_data['bboxes'][0]['x1'], img_data['bboxes'][0]['x1'])
        self.assertNotEqual(prepared_data['bboxes'][0]['y1'], img_data['bboxes'][0]['y1'])
        self.assertEqual(prepared_data['original_height'], w)  # Rot 90
        self.assertEqual(prepared_data['original_width'], h)  # Rot 90
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        #
        mock_np_random_choice.return_value = [180]
        prepared_data = model._generate_images_with_bboxes(img_data_simple, horizontal_flip=False, vertical_flip=False, rot_90=True)
        self.assertEqual(prepared_data['img'].shape[0], resize_h)
        self.assertEqual(prepared_data['img'].shape[1], resize_w)
        self.assertEqual(prepared_data['bboxes'], [])
        self.assertEqual(prepared_data['original_height'], h)  # Rot 180
        self.assertEqual(prepared_data['original_width'], w)  # Rot 180
        self.assertEqual(prepared_data['resized_height'], resize_h)
        self.assertEqual(prepared_data['resized_width'], resize_w)
        remove_dir(model_dir)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test11_model_keras_faster_rcnn_get_preprocess_input(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_preprocess_input'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        im_path = os.path.join(os.getcwd(), 'test_data', 'apple_36.jpg')
        with open(im_path, 'rb') as f:
            img = Image.open(io.BytesIO(f.read()))
            if img.mode != 'RGB':
                img = img.convert('RGB')
            # Convert to array
            img = np.asarray(img)
        h, w = img.shape[:2]  # H > W

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100)
        preprocess_input = model._get_preprocess_input()
        new_img = preprocess_input(img)
        new_h, new_w = new_img.shape[:2]
        self.assertEqual(new_w, 100)
        remove_dir(model_dir)
        #
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=350)
        preprocess_input = model._get_preprocess_input()
        new_img = preprocess_input(img)
        new_h, new_w = new_img.shape[:2]
        self.assertEqual(new_w, 350)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test12_model_keras_faster_rcnn_fit_object_detector(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._fit_object_detector'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model._fit_object_detector(df_data, df_valid=None, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['apple', 'banana', 'orange'])
        self.assertTrue(model.model_rpn._is_compiled)
        self.assertTrue(model.model_classifier._is_compiled)
        self.assertTrue(model.model._is_compiled)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertAlmostEqual(0.013, float(model.model_rpn.optimizer._decayed_lr(tf.float32).numpy()))
        self.assertAlmostEqual(0.014, float(model.model_classifier.optimizer._decayed_lr(tf.float32).numpy()))
        remove_dir(model_dir)
        # With valid & shuffle à False
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model._fit_object_detector(df_data, df_valid=df_data, with_shuffle=False)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(sorted(model.list_classes), ['apple', 'banana', 'orange'])
        self.assertTrue(model.model_rpn._is_compiled)
        self.assertTrue(model.model_classifier._is_compiled)
        self.assertTrue(model.model._is_compiled)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertAlmostEqual(0.013, float(model.model_rpn.optimizer._decayed_lr(tf.float32).numpy()))
        self.assertAlmostEqual(0.014, float(model.model_classifier.optimizer._decayed_lr(tf.float32).numpy()))
        remove_dir(model_dir)

        # Test continue training
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        model.fit(df_data, df_valid=df_data, with_shuffle=True)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        model.save()
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # 2nd fit
        model.fit(df_data, df_valid=df_data, with_shuffle=True)  # We fit again with the same data, not important
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 2)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        # We do not save on purpose
        model_dir_2 = model.model_dir
        self.assertNotEqual(model_dir, model_dir_2)
        # 3rd fit
        model.fit(df_data, df_valid=df_data, with_shuffle=True)  # We fit again with the same data, not important
        model.save()
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 3)
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_1.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'configurations_fit_2.json')))
        model_dir_3 = model.model_dir
        self.assertNotEqual(model_dir_2, model_dir_3)
        # Assertion errors (bad classes)
        old_dict_classes = model.dict_classes.copy()
        model.dict_classes = {'a': 0, 'b': 1}
        with self.assertRaises(AssertionError):
            model.fit(df_data, df_valid=df_data, with_shuffle=True)
        model.dict_classes = old_dict_classes
        model.list_classes = ['a', 'b']
        with self.assertRaises(AssertionError):
            model.fit(df_data, df_valid=df_data, with_shuffle=True)
        remove_dir(model_dir)
        remove_dir(model_dir_2)
        remove_dir(model_dir_3)

        # Manage errors
        df_data['bboxes'][0][0]['class'] = 'bg'
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10,
                                                   lr_rpn_trainable_true=0.011, lr_classifier_trainable_true=0.012, lr_rpn_trainable_false=0.013,
                                                   lr_classifier_trainable_false=0.014, keras_params={'decay_rpn': 0.0, 'decay_classifier': 0.0})
        with self.assertRaises(ValueError):
            model.fit(df_data, df_valid=df_data, with_shuffle=True)

        # Clean
        remove_dir(model_dir)


    # We do not test _fit_object_detector_RPN & _fit_object_detector_classifier, already done via _fit_object_detector

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test13_model_keras_faster_rcnn_get_callbacks(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_callbacks'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        callbacks = model._get_callbacks(patience=12)
        callbacks_types = [type(_) for _ in callbacks]
        self.assertTrue(tf.keras.callbacks.EarlyStopping in callbacks_types)
        self.assertTrue(ModelCheckpointAll in callbacks_types)
        self.assertTrue(tf.keras.callbacks.CSVLogger in callbacks_types)
        self.assertTrue(tf.keras.callbacks.TerminateOnNaN in callbacks_types)
        checkpoint = callbacks[callbacks_types.index(ModelCheckpointAll)]
        csv_logger = callbacks[callbacks_types.index(tf.keras.callbacks.CSVLogger)]
        early_stopping = callbacks[callbacks_types.index(tf.keras.callbacks.EarlyStopping)]
        self.assertEqual(checkpoint.filepath, os.path.join(model.model_dir, 'best.hdf5'))
        self.assertEqual(csv_logger.filename, os.path.join(model.model_dir, 'logger.csv'))
        self.assertEqual(early_stopping.patience, 12)

        # level save 'LOW'
        model.level_save = 'LOW'
        callbacks = model._get_callbacks(patience=14)
        callbacks_types = [type(_) for _ in callbacks]
        self.assertTrue(tf.keras.callbacks.EarlyStopping in callbacks_types)
        self.assertFalse(ModelCheckpointAll in callbacks_types)
        self.assertTrue(tf.keras.callbacks.CSVLogger in callbacks_types)
        self.assertTrue(tf.keras.callbacks.TerminateOnNaN in callbacks_types)
        csv_logger = callbacks[callbacks_types.index(tf.keras.callbacks.CSVLogger)]
        early_stopping = callbacks[callbacks_types.index(tf.keras.callbacks.EarlyStopping)]
        self.assertEqual(csv_logger.filename, os.path.join(model.model_dir, 'logger.csv'))
        self.assertEqual(early_stopping.patience, 14)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test14_model_keras_faster_rcnn_get_learning_rate_scheduler(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._get_learning_rate_scheduler'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        self.assertEqual(model._get_learning_rate_scheduler(), None)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test15_model_keras_faster_rcnn_predict_object_detector(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector._predict_object_detector'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        model.fit(df_data, df_valid=df_data)
        preds = model.predict(df_data)
        self.assertEqual(len(preds), df_data.shape[0])
        # We can't test lots of things ...
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
            model.predict(df_data)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test16_model_keras_faster_rcnn_rcnn_plot_metrics_and_loss(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector._plot_metrics_and_loss'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        class FitHistory():
            def __init__(self, history) -> None:
                self.history = history
        fit_history = FitHistory({
            'loss': [1, 2, 3], 'val_loss': [1, 2, 3],
            'rpn_class_loss': [1, 2, 3], 'val_rpn_class_loss': [1, 2, 3],
            'rpn_regr_loss': [1, 2, 3], 'val_rpn_regr_loss': [1, 2, 3],
            'rpn_class_accuracy': [1, 2, 3], 'val_rpn_class_accuracy': [1, 2, 3],
            'dense_class_loss': [1, 2, 3], 'val_dense_class_loss': [1, 2, 3],
            'dense_regr_loss': [1, 2, 3], 'val_dense_regr_loss': [1, 2, 3],
            'dense_class_accuracy': [1, 2, 3], 'val_dense_class_accuracy': [1, 2, 3],
        })
        fit_history_small = FitHistory({
            'loss': [1, 2, 3], 'val_loss': [1, 2, 3],
            'rpn_regr_loss': [1, 2, 3], 'val_rpn_regr_loss': [1, 2, 3],
        })
        #
        targets_name =  {f'{model_type}_{trainable}': [
                            f'loss_{model_type}_trainable_{trainable}',
                            f'loss_class_{model_type}_trainable_{trainable}',
                            f'loss_regr_{model_type}_trainable_{trainable}',
                            f'accuracy_class_{model_type}_trainable_{trainable}',
                        ] for model_type in ['rpn', 'classifier'] for trainable in [False, True]}

        # Set a model
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir)
        plots_path = os.path.join(model.model_dir, 'plots')

        # Nominal case
        model._plot_metrics_and_loss(fit_history, model_type='rpn', trainable=True)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['rpn_True']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)
        #
        model._plot_metrics_and_loss(fit_history, model_type='classifier', trainable=True)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['classifier_True']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)
        #
        model._plot_metrics_and_loss(fit_history, model_type='rpn', trainable=False)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['rpn_False']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)
        #
        model._plot_metrics_and_loss(fit_history, model_type='classifier', trainable=False)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 4)
        for filename in targets_name['classifier_False']:
            self.assertTrue(os.path.exists(os.path.join(plots_path, f"{filename}.jpeg")))
        remove_dir(plots_path)

        # Less metrics
        model._plot_metrics_and_loss(fit_history_small, model_type='rpn', trainable=True)
        list_files = os.listdir(plots_path)
        self.assertEqual(len(list_files), 2)
        remove_dir(plots_path)

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test17_model_keras_faster_rcnn_save(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector.save'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()

        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        model.fit(df_data)
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"preprocess_input.pkl")))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'keras')
        self.assertTrue('batch_size' in configs.keys())
        self.assertTrue('epochs' in configs.keys())
        self.assertTrue('validation_split' in configs.keys())
        self.assertTrue('patience' in configs.keys())
        self.assertTrue('width' in configs.keys())
        self.assertTrue('height' in configs.keys())
        self.assertTrue('depth' in configs.keys())
        self.assertTrue('color_mode' in configs.keys())
        self.assertTrue('in_memory' in configs.keys())
        self.assertTrue('data_augmentation_params' in configs.keys())
        self.assertTrue('nb_train_generator_images_to_save' in configs.keys())
        self.assertTrue('keras_params' in configs.keys())
        self.assertTrue('_get_model' in configs.keys())
        self.assertTrue('_get_learning_rate_scheduler' in configs.keys())
        self.assertTrue('custom_objects' in configs.keys())
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        self.assertTrue('vgg_filename' in configs.keys())
        self.assertTrue('shared_model_subsampling' in configs.keys())
        self.assertTrue('anchor_box_sizes' in configs.keys())
        self.assertTrue('anchor_box_ratios' in configs.keys())
        self.assertTrue('nb_anchors' in configs.keys())
        self.assertTrue('list_anchors' in configs.keys())
        self.assertTrue('img_min_side_size' in configs.keys())
        self.assertTrue('pool_resize_classifier' in configs.keys())
        self.assertTrue('rpn_regr_scaling' in configs.keys())
        self.assertTrue('classifier_regr_scaling' in configs.keys())
        self.assertTrue('rpn_min_overlap' in configs.keys())
        self.assertTrue('rpn_max_overlap' in configs.keys())
        self.assertTrue('rpn_restrict_num_regions' in configs.keys())
        self.assertTrue('nb_rois_classifier' in configs.keys())
        self.assertTrue('roi_nms_overlap_threshold' in configs.keys())
        self.assertTrue('nms_max_boxes' in configs.keys())
        self.assertTrue('classifier_min_overlap' in configs.keys())
        self.assertTrue('classifier_max_overlap' in configs.keys())
        self.assertTrue('pred_bbox_proba_threshold' in configs.keys())
        self.assertTrue('pred_nms_overlap_threshold' in configs.keys())
        self.assertTrue('batch_size_rpn_trainable_true' in configs.keys())
        self.assertTrue('batch_size_classifier_trainable_true' in configs.keys())
        self.assertTrue('batch_size_rpn_trainable_false' in configs.keys())
        self.assertTrue('batch_size_classifier_trainable_false' in configs.keys())
        self.assertTrue('epochs_rpn_trainable_true' in configs.keys())
        self.assertTrue('epochs_classifier_trainable_true' in configs.keys())
        self.assertTrue('epochs_rpn_trainable_false' in configs.keys())
        self.assertTrue('epochs_classifier_trainable_false' in configs.keys())
        self.assertTrue('patience_rpn_trainable_true' in configs.keys())
        self.assertTrue('patience_classifier_trainable_true' in configs.keys())
        self.assertTrue('patience_rpn_trainable_false' in configs.keys())
        self.assertTrue('patience_classifier_trainable_false' in configs.keys())
        self.assertTrue('_add_rpn_layers' in configs.keys())
        self.assertTrue('_add_classifier_layers' in configs.keys())
        self.assertTrue('_compile_model_rpn' in configs.keys())
        self.assertTrue('_compile_model_classifier' in configs.keys())
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test18_model_keras_faster_rcnn_reload_model(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector.reload_model'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()


        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        model.fit(df_data)
        model.save()
        # Reload keras
        hdf5_path = os.path.join(model.model_dir, 'best.hdf5')
        reloaded_model = model.reload_model(hdf5_path)
        # We can't really test the predictions because chances are high that no bboxes are predicted

        # Clean
        remove_dir(model_dir)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test19_model_keras_faster_rcnn_reload_from_standalone(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn_rcnn.ModelDetectronFasterRcnnObjectDetector.reload_from_standalone'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()


        # Nominal case
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10)
        fit_params = model.fit(df_data)
        model.save()
        # Reload
        conf_path = os.path.join(model.model_dir, "configurations.json")
        hdf5_path = os.path.join(model.model_dir, "best.hdf5")
        preprocess_input_path = os.path.join(model.model_dir, "preprocess_input.pkl")
        new_model = ModelKerasFasterRcnnObjectDetector()
        new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
                                         preprocess_input_path=preprocess_input_path)
        # Tests
        self.assertEqual(model.model_name, new_model.model_name)
        self.assertEqual(model.model_type, new_model.model_type)
        self.assertEqual(model.trained, new_model.trained)
        self.assertEqual(model.nb_fit, new_model.nb_fit)
        self.assertEqual(model.level_save, new_model.level_save)
        self.assertEqual(model.list_classes, new_model.list_classes)
        self.assertEqual(model.dict_classes, new_model.dict_classes)
        self.assertEqual(model.batch_size, new_model.batch_size)
        self.assertEqual(model.epochs, new_model.epochs)
        self.assertEqual(model.validation_split, new_model.validation_split)
        self.assertEqual(model.patience, new_model.patience)
        self.assertEqual(model.width, new_model.width)
        self.assertEqual(model.height, new_model.height)
        self.assertEqual(model.depth, new_model.depth)
        self.assertEqual(model.color_mode, new_model.color_mode)
        self.assertEqual(model.in_memory, new_model.in_memory)
        self.assertEqual(model.data_augmentation_params, new_model.data_augmentation_params)
        self.assertEqual(model.nb_train_generator_images_to_save, new_model.nb_train_generator_images_to_save)
        self.assertEqual(model.vgg_filename, new_model.vgg_filename)
        self.assertEqual(model.vgg_path, new_model.vgg_path)
        self.assertEqual(model.shared_model_subsampling, new_model.shared_model_subsampling)
        self.assertEqual(model.anchor_box_sizes, new_model.anchor_box_sizes)
        self.assertEqual(model.anchor_box_ratios, new_model.anchor_box_ratios)
        self.assertEqual(model.nb_anchors, new_model.nb_anchors)
        self.assertEqual(model.list_anchors, new_model.list_anchors)
        self.assertEqual(model.img_min_side_size, new_model.img_min_side_size)
        self.assertEqual(model.pool_resize_classifier, new_model.pool_resize_classifier)
        self.assertEqual(model.rpn_regr_scaling, new_model.rpn_regr_scaling)
        self.assertEqual(model.classifier_regr_scaling, new_model.classifier_regr_scaling)
        self.assertEqual(model.rpn_min_overlap, new_model.rpn_min_overlap)
        self.assertEqual(model.rpn_max_overlap, new_model.rpn_max_overlap)
        self.assertEqual(model.rpn_restrict_num_regions, new_model.rpn_restrict_num_regions)
        self.assertEqual(model.nb_rois_classifier, new_model.nb_rois_classifier)
        self.assertEqual(model.roi_nms_overlap_threshold, new_model.roi_nms_overlap_threshold)
        self.assertEqual(model.nms_max_boxes, new_model.nms_max_boxes)
        self.assertEqual(model.classifier_min_overlap, new_model.classifier_min_overlap)
        self.assertEqual(model.classifier_max_overlap, new_model.classifier_max_overlap)
        self.assertEqual(model.pred_bbox_proba_threshold, new_model.pred_bbox_proba_threshold)
        self.assertEqual(model.pred_nms_overlap_threshold, new_model.pred_nms_overlap_threshold)
        self.assertEqual(model.batch_size_rpn_trainable_true, new_model.batch_size_rpn_trainable_true)
        self.assertEqual(model.batch_size_classifier_trainable_true, new_model.batch_size_classifier_trainable_true)
        self.assertEqual(model.batch_size_classifier_trainable_true, new_model.batch_size_classifier_trainable_true)
        self.assertEqual(model.batch_size_rpn_trainable_false, new_model.batch_size_rpn_trainable_false)
        self.assertEqual(model.batch_size_rpn_trainable_false, new_model.batch_size_rpn_trainable_false)
        self.assertEqual(model.batch_size_classifier_trainable_false, new_model.batch_size_classifier_trainable_false)
        self.assertEqual(model.batch_size_classifier_trainable_false, new_model.batch_size_classifier_trainable_false)
        self.assertEqual(model.epochs_rpn_trainable_true, new_model.epochs_rpn_trainable_true)
        self.assertEqual(model.epochs_classifier_trainable_true, new_model.epochs_classifier_trainable_true)
        self.assertEqual(model.epochs_classifier_trainable_true, new_model.epochs_classifier_trainable_true)
        self.assertEqual(model.epochs_rpn_trainable_false, new_model.epochs_rpn_trainable_false)
        self.assertEqual(model.epochs_classifier_trainable_false, new_model.epochs_classifier_trainable_false)
        self.assertEqual(model.epochs_classifier_trainable_false, new_model.epochs_classifier_trainable_false)
        self.assertEqual(model.patience_rpn_trainable_true, new_model.patience_rpn_trainable_true)
        self.assertEqual(model.patience_classifier_trainable_true, new_model.patience_classifier_trainable_true)
        self.assertEqual(model.patience_classifier_trainable_true, new_model.patience_classifier_trainable_true)
        self.assertEqual(model.patience_rpn_trainable_false, new_model.patience_rpn_trainable_false)
        self.assertEqual(model.patience_classifier_trainable_false, new_model.patience_classifier_trainable_false)
        self.assertEqual(model.patience_classifier_trainable_false, new_model.patience_classifier_trainable_false)
        self.assertEqual(model.keras_params, new_model.keras_params)
        # self.assertEqual(model.custom_objects, new_model.custom_objects)  # Do not work because functions are wrapped ...
        # We can't really test the pipeline so we test predictions
        self.assertEqual(model.predict(df_data), new_model.predict(df_data))  # Useful ?
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)


        ############################################
        # Errors
        ############################################

        # with self.assertRaises(FileNotFoundError):
        #     new_model = ModelCnnClassifier()
        #     new_model.reload_from_standalone(configuration_path='toto.json', hdf5_path=hdf5_path,
        #                                      preprocess_input_path=preprocess_input_path)
        # with self.assertRaises(FileNotFoundError):
        #     new_model = ModelCnnClassifier()
        #     new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path='toto.pkl',
        #                                      preprocess_input_path=preprocess_input_path)
        # with self.assertRaises(FileNotFoundError):
        #     new_model = ModelCnnClassifier()
        #     new_model.reload_from_standalone(configuration_path=conf_path, hdf5_path=hdf5_path,
        #                                      preprocess_input_path='toto.pkl')

        # Clean
        remove_dir(model_dir)


class CustomGeneratorRpnTests(unittest.TestCase):
    '''Main class to test model_keras_faster_rcnn.CustomGeneratorRpn'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector', side_effect=MockModelKerasFasterRcnnObjectDetector)
    def test01_custom_generator_rpn_init(self, mock_model):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorRpn.__init__'''

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        fake_model = mock_model()

        # Init., test all parameters
        generator = CustomGeneratorRpn(img_data_list, batch_size=5, shuffle=True, seed=42, model=fake_model,
                                       horizontal_flip=False, vertical_flip=False, rot_90=False, data_type='train',
                                       with_img_data=False)
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, 42)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.is_test, False)
        # Other params
        generator = CustomGeneratorRpn(img_data_list, batch_size=10, shuffle=False, seed=64, model=fake_model,
                                       horizontal_flip=True, vertical_flip=True, rot_90=True, data_type='test',
                                       with_img_data=True)
        self.assertEqual(generator.batch_size, 1)  # test data, force batch size 1
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, 64)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.with_img_data, True)
        self.assertEqual(generator.is_test, True)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.utils_object_detectors.get_rpn_targets', side_effect=mock_get_rpn_targets)
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector', side_effect=MockModelKerasFasterRcnnObjectDetector)
    def test02_custom_generator_rpn_get_batches_of_transformed_samples(self, mock_model, mock_rpn_targets):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorRpn._get_batches_of_transformed_samples'''

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        fake_model = mock_model()

        # Nominal case
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='train', with_img_data=False)
        X, Y = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(sorted(list(Y.keys())), sorted(['rpn_class', 'rpn_regr']))
        # with_img_data = True
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='train', with_img_data=True)
        X, Y, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(sorted(list(Y.keys())), sorted(['rpn_class', 'rpn_regr']))
        self.assertEqual(len(batch_data), 2)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())
        # data_type == 'test'
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='test', with_img_data=False)
        X = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        # data_type == 'test' & with_img_data == True
        generator = CustomGeneratorRpn(img_data_list, batch_size=2, shuffle=True, seed=42, model=fake_model,
                                       data_type='test', with_img_data=True)
        X, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(list(X.keys()), ['input_img'])
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(len(batch_data), 1)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())


class CustomGeneratorClassifierTests(unittest.TestCase):
    '''Main class to test model_keras_faster_rcnn.CustomGeneratorClassifier'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.clone_model', side_effect=mock_clone_model_fn)
    @patch('{{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.ModelKerasFasterRcnnObjectDetector', side_effect=MockModelKerasFasterRcnnObjectDetector)
    def test01_custom_generator_classifier_init(self, mock_model, mock_clone_model):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorClassifier.__init__'''

        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        fake_model = mock_model()

        # Init., test all parameters
        generator = CustomGeneratorClassifier(img_data_list, batch_size=5, shuffle=True, seed=42, model=fake_model, shared_model_trainable=False,
                                              horizontal_flip=False, vertical_flip=False, rot_90=False, data_type='train', with_img_data=False)
        self.assertEqual(generator.batch_size, 5)
        self.assertEqual(generator.shuffle, True)
        self.assertEqual(generator.seed, 42)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, False)
        self.assertEqual(generator.vertical_flip, False)
        self.assertEqual(generator.rot_90, False)
        self.assertEqual(generator.with_img_data, False)
        self.assertEqual(generator.is_test, False)
        self.assertTrue(generator.rpn_clone is None)
        # Other params
        generator = CustomGeneratorClassifier(img_data_list, batch_size=10, shuffle=False, seed=64, model=fake_model, shared_model_trainable=True,
                                              horizontal_flip=True, vertical_flip=True, rot_90=True, data_type='test', with_img_data=True)
        self.assertEqual(generator.batch_size, 1)  # test data, force batch size 1
        self.assertEqual(generator.shuffle, False)
        self.assertEqual(generator.seed, 64)
        self.assertEqual(generator.img_data_list, img_data_list)
        self.assertEqual(generator.model, fake_model)
        self.assertEqual(generator.horizontal_flip, True)
        self.assertEqual(generator.vertical_flip, True)
        self.assertEqual(generator.rot_90, True)
        self.assertEqual(generator.with_img_data, True)
        self.assertEqual(generator.is_test, True)
        self.assertFalse(generator.rpn_clone is None)

    # We do not mock anything, too complex
    @unittest.skip('This test should pass but is skipped to avoid OOM in the deployment CI')
    def test02_custom_generator_classifier_get_batches_of_transformed_samples(self):
        '''Test of {{package_name}}.models_training.object_detectors.model_keras_faster_rcnn.CustomGeneratorClassifier._get_batches_of_transformed_samples'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars & model
        data_path = os.path.join(os.getcwd(), 'test_data', 'test_data_object_detection', 'fruits')
        path_list, bboxes_list, _ = utils.read_folder_object_detection(data_path, images_ext=('.jpg', '.jpeg', '.png'))
        df_data = pd.DataFrame({'file_path': path_list, 'bboxes': bboxes_list}).copy()
        img_data_list = []
        for filepath, bboxes in zip(path_list, bboxes_list):
            img_data_list.append({'file_path': filepath, 'bboxes': bboxes})
        model = ModelKerasFasterRcnnObjectDetector(model_dir=model_dir, img_min_side_size=100, epochs=1, batch_size=2, nms_max_boxes=10, nb_rois_classifier=3)
        model.fit(df_data)

        # Nominal case
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=False, data_type='train', with_img_data=False)
        X, Y = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 2)
        self.assertEqual(X['input_rois'].shape[1], model.nb_rois_classifier)
        self.assertEqual(X['input_rois'].shape[2], 4)
        self.assertEqual(sorted(list(Y.keys())), sorted(['dense_class', 'dense_regr']))
        self.assertEqual(Y['dense_class'].shape[0], 2)
        self.assertEqual(Y['dense_class'].shape[1], model.nb_rois_classifier)
        self.assertEqual(Y['dense_regr'].shape[0], 2)
        self.assertEqual(Y['dense_regr'].shape[1], model.nb_rois_classifier)
        # with_img_data = True & shared_model_trainable à True
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=True, data_type='train', with_img_data=True)
        X, Y, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([3, 1]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 2)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 2)
        self.assertEqual(X['input_rois'].shape[1], model.nb_rois_classifier)
        self.assertEqual(X['input_rois'].shape[2], 4)
        self.assertEqual(sorted(list(Y.keys())), sorted(['dense_class', 'dense_regr']))
        self.assertEqual(Y['dense_class'].shape[0], 2)
        self.assertEqual(Y['dense_class'].shape[1], model.nb_rois_classifier)
        self.assertEqual(Y['dense_regr'].shape[0], 2)
        self.assertEqual(Y['dense_regr'].shape[1], model.nb_rois_classifier)
        self.assertEqual(len(batch_data), 2)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())
        # data_type == 'test'
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=True, data_type='test', with_img_data=False)
        X = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 1)
        # self.assertEqual(X['input_rois'].shape[1], ???) -> We do not know the number of ROIs returned for the test
        self.assertEqual(X['input_rois'].shape[2], 4)
        # data_type == 'test' & with_img_data == True
        generator = CustomGeneratorClassifier(img_data_list, batch_size=2, shuffle=True, seed=42, model=model,
                                              shared_model_trainable=True, data_type='test', with_img_data=True)
        X, batch_data = generator._get_batches_of_transformed_samples(index_array=np.array([2]))
        self.assertEqual(sorted(list(X.keys())), sorted(['input_img', 'input_rois']))
        self.assertEqual(X['input_img'].shape[0], 1)
        self.assertEqual(X['input_img'].shape[-1], 3)
        self.assertEqual(X['input_rois'].shape[0], 1)
        # self.assertEqual(X['input_rois'].shape[1], ???) -> We do not know the number of ROIs returned for the test
        self.assertEqual(X['input_rois'].shape[2], 4)
        self.assertEqual(len(batch_data), 1)
        self.assertTrue('batch_width' in batch_data[0].keys())
        self.assertTrue('batch_height' in batch_data[0].keys())


# We do not test ModelCheckpointAll


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Outliers detection
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import math
import cmath
import logging
import numpy as np
import pandas as pd
from typing import Union
import scipy.integrate as integrate
from sklearn.ensemble import IsolationForest
from sklearn.neighbors import LocalOutlierFactor

logger = logging.getLogger(__name__)


def check_for_outliers(X: Union[pd.DataFrame, np.ndarray], n_estimators: int = 100, n_neighbors: int = 20) -> np.ndarray:
    '''Agreggates two results of outliers detection and warns the user if some were detected

    Args :
        X (np.ndarray, pd.DataFrame): Shape = [n_samples, n_features]
    Kwargs:
        n_estimators (int): number of estimators for the IsolationForest
            If 0, do not IsolationForest
        n_neighbors (int): number of neighbors for the LocalOutlierFactor
            If 0, do not use LocalOutlierFactor
    Raises:
        ValueError: If n_estimators < 0
        ValueError: If n_neighbors < 0
        ValueError: If both n_estimators and n_neighbors are equal to 0
    Returns:
        outliers (np.ndarray): 1d array of n_samples containing -1 if outlier, 1 otherwise
    '''
    # Manage errors
    if n_estimators < 0:
        raise ValueError("n_estimators must be positive")
    if n_neighbors < 0:
        raise ValueError("n_neighbors must be positive")
    if n_estimators + n_neighbors == 0:
        raise ValueError("n_neighbors and n_estimators can't both be equal to 0")

    # Init. outliers (1 = not an outlier, -1 = outlier)
    outliers = np.ones(X.shape[0], dtype=int)

    # Get outliers from IsolationForest
    if not n_estimators == 0:
        run_forest = IsolationForest(n_estimators=n_estimators)
        outliers |= run_forest.fit_predict(X)  # In-place union
    else:
        logger.info("IsolationForest is skipped (n_estimators == 0)")

    # Get outliers from LocalOutlierFactor
    if not n_neighbors == 0:
        lof = LocalOutlierFactor(n_neighbors=n_neighbors)
        outliers |= lof.fit_predict(X)  # In-place union
    else:
        logger.info("LocalOutlierFactor is skipped (n_neighbors == 0)")

    # Logger
    if int(cmath.exp(1j*integrate.quad(lambda x: math.sqrt(1 - pow(x, 2)), -1, 1)[0]*2).real) in outliers:
        logger.warning("The dataset seems to contain outliers at indices:")
        logger.warning(", ".join(str(v) for v in list(np.where(outliers==-1)[0])))

    # Return outliers
    return outliers


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Utils - Tools for training
# Copyright (C) <2018-2021>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Functions :
# - normal_split -> Splits a DataFrame into train and test sets
# - stratified_split -> Splits a DataFrame into train and test sets - Stratified strategy
# - remove_small_classes -> Deletes under-represented classes
# - display_train_test_shape -> Displays the size of a train/test split
# - load_model -> Loads a model from a path
# - predict -> Gets predictions on a dataset
# - predict_with_proba -> Gets predictions with probabilities on a dataset


import os
import json
import math
import tempfile
import dill as pickle
import logging
import numpy as np
import pandas as pd
from PIL import Image
from typing import Union, Tuple, Any, List

from sklearn.model_selection import train_test_split

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess


# Get logger
logger = logging.getLogger(__name__)


def normal_split(df: pd.DataFrame, test_size: float = 0.25, seed: Union[int, None] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''Splits a DataFrame into train and test sets

    Args:
        df (pd.DataFrame): Dataframe containing the data
    Kwargs:
        test_size (float): Proportion representing the size of the expected test set
        seed (int): random seed
    Raises:
        ValueError: If the object test_size is not between 0 and 1
    Returns:
        DataFrame: Train dataframe
        DataFrame: Test dataframe
    '''
    if not 0 <= test_size <= 1:
        raise ValueError('The object test_size must be between 0 and 1')

    # Normal split
    logger.info("Normal split")
    df_train, df_test = train_test_split(df, test_size=test_size, random_state=seed)

    # Display
    display_train_test_shape(df_train, df_test, df_shape=df.shape[0])

    # Return
    return df_train, df_test


def stratified_split(df: pd.DataFrame, col: Union[str, int], test_size: float = 0.25,
                     seed: Union[int, None] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''Splits a DataFrame into train and test sets - Stratified strategy

    Args:
        df (pd.DataFrame): Dataframe containing the data
        col (str or int): column on which to do the stratified split
    Kwargs:
        test_size (float): Proportion representing the size of the expected test set
        seed (int): Random seed
    Raises:
        ValueError: If the object test_size is not between 0 and 1
    Returns:
        DataFrame: Train dataframe
        DataFrame: Test dataframe
    '''
    if not 0 <= test_size <= 1:
        raise ValueError('The object test_size must be between 0 and 1')

    # Stratified split
    logger.info("Stratified split")
    df = remove_small_classes(df, col, min_rows=math.ceil(1 / test_size))  # minimum lines number per category to split
    df_train, df_test = train_test_split(df, stratify=df[col], test_size=test_size, random_state=seed)

    # Display
    display_train_test_shape(df_train, df_test, df_shape=df.shape[0])

    # Return
    return df_train, df_test


def remove_small_classes(df: pd.DataFrame, col: Union[str, int], min_rows: int = 2) -> pd.DataFrame:
    '''Deletes the classes with small numbers of elements

    Args:
        df (pd.DataFrame): Dataframe containing the data
        col (str | int): Columns containing the classes
    Kwargs:
        min_rows (int): Minimal number of lines in the training set (default: 2)
    Raises:
        ValueError: If the object min_rows is not positive
    Returns:
        pd.DataFrame: New dataset
    '''
    if min_rows < 1:
        raise ValueError("The object min_rows must be positive")

    # Looking for classes with less than min_rows lines
    v_count = df[col].value_counts()
    classes_to_remove = list(v_count[v_count < min_rows].index.values)
    for cl in classes_to_remove:
        logger.warning(f"/!\\ /!\\ /!\\ Class {cl} has less than {min_rows} lines in the training set.")
        logger.warning(f"/!\\ /!\\ /!\\ This class is automatically removed from the dataset.")
    return df[~df[col].isin(classes_to_remove)]


def display_train_test_shape(df_train: pd.DataFrame, df_test: pd.DataFrame, df_shape: Union[int, None] = None) -> None:
    '''Displays the size of a train/test split

    Args:
        df_train (pd.DataFrame): Train dataset
        df_test (pd.DataFrame): Test dataset
    Kwargs:
        df_shape (int): Size of the initial dataset
    Raises:
        ValueError: If the object df_shape is not positive
    '''
    if df_shape is not None and df_shape < 1:
        raise ValueError("The object df_shape must be positive")

    # Process
    if df_shape is None:
        df_shape = df_train.shape[0] + df_test.shape[0]
    logger.info(f"There are {df_train.shape[0]} lines in the train dataset and {df_test.shape[0]} in the test dataset.")
    logger.info(f"{round(100 * df_train.shape[0] / df_shape, 2)}% of data are in the train set")
    logger.info(f"{round(100 * df_test.shape[0] / df_shape, 2)}% of data are in the test set")


def load_model(model_dir: str, is_path: bool = False) -> Tuple[Any, dict]:
    '''Loads a model from a path

    Args:
        model_dir (str): Name of the folder containing the model (e.g. model_autres_2019_11_07-13_43_19)
    Kwargs:
        is_path (bool): If folder path instead of name (permits to load model from elsewhere)
    Raises:
        FileNotFoundError: If the folder model_dir does not exist
    Returns:
        ?: Model
        dict: Model configurations
    '''

    # Find model path
    if not is_path:
        models_dir = utils.get_models_path()
        model_path = None
        for path, subdirs, files in os.walk(models_dir):
            for name in subdirs:
                if name == model_dir:
                    model_path = os.path.join(path, name)
        if model_path is None:
            raise FileNotFoundError(f"Can't find model {model_dir}")
    else:
        model_path = model_dir
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"Can't find model {model_path} (considered as a path)")


    # Get configs
    configuration_path = os.path.join(model_path, 'configurations.json')
    with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
        configs = json.load(f)
    # Can't set int as keys in json, so need to cast it after reloading
    # dict_classes keys are always ints
    if 'dict_classes' in configs.keys() and configs['dict_classes'] is not None:
        configs['dict_classes'] = {int(k): v for k, v in configs['dict_classes'].items()}

    # Load model
    pkl_path = os.path.join(model_path, f"{configs['model_name']}.pkl")
    with open(pkl_path, 'rb') as f:
        model = pickle.load(f)

    # Change model_dir if diff
    if model_path != model.model_dir:
        model.model_dir = model_path
        configs['model_dir'] = model_path

    # Load specifics
    hdf5_path = os.path.join(model_path, 'best.hdf5')

    # Check for keras model
    if os.path.exists(hdf5_path):
        # Si on a définit une fonction spéciale (faster RCNN), on l'utilise
        if hasattr(model, 'reload_models_from_hdf5'):
            model.reload_models_from_hdf5(hdf5_path)
        else:
            model.model = model.reload_model(hdf5_path)

    # Display if GPU is being used
    model.display_if_gpu_activated()

    # Return model & configs
    return model, configs


def predict(data_input: Union[str, List[str], np.ndarray, pd.DataFrame], model, model_conf: dict,
            return_proba: bool = False, **kwargs) -> Union[str, List[str], np.ndarray]:
    '''Gets predictions of a model on images

    Args:
        data_input (str | list<str> | np.ndarray): New content to be predicted
            - str: abs. path to an image
            - list<str>: list of abs. path to an image
            - np.ndarray: an already loaded image
                Possibility to have several images if 4 dim (i.e (nb_images, width, height, channels))
            - pd.DataFrame: Dataframe with a column file_path (abs. paths to images)
        model (ModelClass): Model to use
        model_conf (dict): Model configurations
    Kwargs:
        return_proba (bool): If probabilities must be return instead
    Raises:
        NotImplementedError: If model is object detection task
        FileNotFoundError: If the input file does not exist (input type == str)
        FileNotFoundError: If one of the input files does not exist (input type == list)
        ValueError: If the input image format is not compatible (input type == np.ndarray)
        ValueError: If the input array is not compatible (input type == np.ndarray)
        ValueError: If the input DataFrame does not contains a 'file_path' column (input type == pd.DataFrame)
        ValueError: If the input type is not a valid type option
    Returns:
        str, List[str], np.ndarray: predictions or probabilities
            - If return_proba -> np.ndarray (shape depends on number of inputs)
            - Else str or list<str> (depends on number of inputs)
    '''
    # TODO
    # TODO
    # TODO: Make this works with object_detector !!!
    # TODO
    # TODO
    if model.model_type == 'object_detector':
        raise NotImplementedError("`predict` is not yet implemented for object detection task")

    ##############################################
    # Retrieve data - PIL format (list)
    ##############################################

    # Type 1: absolute path
    if isinstance(data_input, str):
        if not os.path.exists(data_input):
            raise FileNotFoundError(f"The file {data_input} does not exist")
        images = [Image.open(data_input)]

    # Type 2: list of absolute paths
    elif isinstance(data_input, list):
        if not all([os.path.exists(_) for _ in data_input]):
            raise FileNotFoundError("At least one of the input path does not exist")
        images = [Image.open(_) for _ in data_input]

    # Type 3: numpy array
    elif isinstance(data_input, np.ndarray):
        # If only one image (shape = 3), exapnd a 4th image
        if len(data_input.shape) == 3:
            data_input = np.expand_dims(data_input, 0)
        # Consider input as image list
        if len(data_input.shape) == 4:
            images = []
            for i in range(data_input.shape[0]):
                np_image = data_input[i]
                # RGB
                if np_image.shape[-1] == 3:
                    images.append(Image.fromarray(np_image, 'RGB'))
                elif np_image.shape[-1] == 4:
                    images.append(Image.fromarray(np_image, 'RGBA'))
                else:
                    raise ValueError(f"Input image format ({np_image.shape}) is not compatible")
        else:
            raise ValueError(f"Input array format ({type(data_input)}) is not valid")

    # Type 4: pd.DataFrame
    elif isinstance(data_input, pd.DataFrame):
        if 'file_path' not in data_input.columns:
            raise ValueError("The input DataFrame does not contains a 'file_path' column (mandatory)")
        file_paths = list(data_input['file_path'].values)
        if not all([os.path.exists(_) for _ in file_paths]):
            raise FileNotFoundError("At least one of the input path does not exist")
        images = [Image.open(_) for _ in file_paths]

    # No solution
    else:
        raise ValueError(f"Input type ({type(data_input)}) is not a valid type option.")


    ##############################################
    # Apply preprocessing
    ##############################################

    # Get preprocessor
    if 'preprocess_str' in model_conf.keys():
        preprocess_str = model_conf['preprocess_str']
    else:
        preprocess_str = "no_preprocess"
    preprocessor = preprocess.get_preprocessor(preprocess_str)

    # Preprocess
    images_preprocessed = preprocessor(images)


    ##############################################
    # Save all preprocessed images in a temporary directory
    ##############################################

    # We'll create a temporary folder to save preprocessed images
    with tempfile.TemporaryDirectory(dir=utils.get_data_path()) as tmp_folder:
        # Save images
        images_path = []
        for i, img in enumerate(images_preprocessed):
            img_path = os.path.join(tmp_folder, f"image_{i}.png")
            img.save(img_path, format='PNG')
            images_path.append(img_path)

        # Get predictions
        df = pd.DataFrame({'file_path': images_path})
        predictions, probas = model.predict_with_proba(df)

    # Getting out of the context, all temporary data is deleted


    ##############################################
    # Return result
    ##############################################

    # Return one element if only one input, else return all
    if return_proba:
        return probas[0] if len(probas) == 1 else probas
    else:
        predictions = model.inverse_transform(predictions)
        return predictions[0] if len(predictions) == 1 else predictions


def predict_with_proba(data_input: Union[str, List[str], np.ndarray, pd.DataFrame], model,
                       model_conf: dict) -> Tuple[Union[str, List[str]], Union[float, List[float]]]:
    '''Gets probabilities predictions of a model on a dataset

    Args:
        data_input (str | list<str> | np.ndarray): New content to be predicted
            - str: abs. path to an image
            - list<str>: list of abs. path to an image
            - np.ndarray: an already loaded image
                Possibility to have several images if 4 dim (i.e (nb_images, width, height, channels))
            - pd.DataFrame: Dataframe with a column file_path (abs. paths to images)
        model (ModelClass): Model to use
        model_conf (dict): Model configurations
    Raises:
        NotImplementedError: If model is object detection task
    Returns:
        str: prediction
        float: probability

        If several elements -> list
    '''
    if model.model_type == 'object_detector':
        raise NotImplementedError("`predict_with_proba` is not yet implemented for object detection task")

    # Get probas
    probas: np.ndarray = predict(data_input, model, model_conf, return_proba=True)

    # Manage cases with only one element
    if len(probas.shape) == 1:
        predictions = model.get_classes_from_proba(np.expand_dims(probas, 0))
        predictions = model.inverse_transform(predictions)[0]
        probas = max(probas)
    # Several elements
    else:
        predictions = model.get_classes_from_proba(probas)
        predictions = model.inverse_transform(predictions)
        probas = list(probas.max(axis=1))

    # Return
    return predictions, probas


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## PyTorch transformer model
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelPyTorchTransformers -> Model for predictions via tranformers pytorch

import os
import json
import pickle
import shutil
import logging
import numpy as np
import pandas as pd
import seaborn as sns
from tqdm import tqdm
from functools import partial
import matplotlib.pyplot as plt
from typing import no_type_check, Union, Callable, Any, Tuple

import torch
import pytorch_lightning as pl
from torch.nn import Softmax, Sigmoid
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data.dataset import TensorDataset
from torch.utils.data.sampler import RandomSampler
from transformers import AdamW, AutoModel, AutoTokenizer, AutoConfig, AutoModelForSequenceClassification

from {{package_name}} import utils
from {{package_name}}.models_training import utils_deep_torch
from {{package_name}}.models_training.model_pytorch import ModelPyTorch

sns.set(style="darkgrid")
tqdm.pandas()


class ModelPyTorchTransformers(ModelPyTorch):
    '''Model for predictions via tranformers pytorch'''

    _default_name = 'model_pytorch_transformers'

    def __init__(self, transformer_name: Union[str, None] = None, max_sequence_length: int = 256,
                 tokenizer_special_tokens: Union[tuple, None] = None,
                 padding: str = "max_length", truncation: bool = True,
                  **kwargs) -> None:
        '''Initialization of the class (see ModelClass & ModelPyTorch for more arguments)

        Args:
            transformer_name (str): Name of the transformer to use
        Kwargs:
            max_sequence_length (int): Maximum number of words per sequence (ie. sentences)
            tokenizer_special_tokens (tuple): Set of "special tokens" for the tokenizer
            padding (str): Tokenizer's padding strategy
            truncation (bool): Tokenizer's padding truncation
        '''
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        # Params
        self.transformer_name = transformer_name
        self.max_sequence_length = max_sequence_length
        if tokenizer_special_tokens is None:
            tokenizer_special_tokens = tuple()
        self.tokenizer_special_tokens = tokenizer_special_tokens
        self.padding = padding
        self.truncation = truncation

        # Retrieve tokenizer
        self.tokenizer: Any = None
        if self.transformer_name is not None:
            self.tokenizer = self._get_tokenizer()
        else:
            self.logger.warning("No transformer name specified. We can't load the tokenizer.")
            self.logger.warning("No training possible if a tokenizer is not loaded (eg. via reload_from_standalone).")

    @utils.data_agnostic_str_to_list
    @utils.trained_needed
    @no_type_check
    def predict_proba(self, x_test, limit_predict_simple: int = 5, **kwargs) -> np.ndarray:
        '''Predicts probabilities on the test dataset

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples]
        Kwargs:
            limit_predict_simple (bool): Maximal number of elements to use a "simple" predict (without dataloaders)
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        # Simplified predictions (faster when considering a small number of data)
        if len(x_test) <= limit_predict_simple:
            input_ids, attention_mask, _ = self._convert_inputs(x_test, None)
            # NEW FIX: convert to correct device
            self.model.convert_network_to_device()  # Convert to gpu if available
            input_ids = self.model.to_device(input_ids)
            attention_mask = self.model.to_device(attention_mask)
            # Switch to eval mode
            self.model.eval()
            # Get predictions
            logits_torch = self.model(input_ids=input_ids, attention_mask=attention_mask)
            logits = np.vstack([logit.cpu() for logit in logits_torch])
            probas = self.model.get_probas_from_logits(logits)
            # Switch back to train mode
            self.model.train()
        # Predictions with DataLoader (slower when considering a small number of data)
        else:
            test_dl = self._get_test_dataloader(self.batch_size, x_test, y_test_dummies=None) # We can change the batch size
            trainer = pl.Trainer(default_root_dir=self.model_dir, checkpoint_callback=False, logger=False,
                                 gpus=1 if torch.cuda.is_available() else 0)
            # Test on x_test
            trainer.test(model=self.model, test_dataloaders=test_dl)
            # Retrieve proba & return it
            probas = self.model.test_probas
        return probas

    def _get_tokenizer(self) -> Any:
        '''Retrieves the tokenizer'''
        # Get tokenizer path
        transformers_path = utils.get_transformers_path()
        transformer_path = os.path.join(transformers_path, self.transformer_name)

        # Check if available locally
        if os.path.exists(transformer_path):
            self.logger.info("Use the local tokenizer")
            final_transformer_localisation = transformer_path
        else:
            self.logger.warning("Can't find the local tokenizer.")
            self.logger.warning("We try to get it from the web.")
            final_transformer_localisation = self.transformer_name

        # Retrieve tokenizer
        tokenizer = AutoTokenizer.from_pretrained(final_transformer_localisation)

        # Add special tokens
        if self.tokenizer_special_tokens:
            tokenizer.add_tokens(self.tokenizer_special_tokens)

        # Return
        return tokenizer

    def _get_transformer(self) -> Any:
        '''Gets a transformer

        Raises:
            ValueError: If list_classes has not been set (ie. fit not called)
        Returns:
            (?): a transformer model
        '''
        if self.list_classes is None:
            raise ValueError("The method '_get_transformer' can't be called if 'fit' has not been called at least once")

        # Set parameters
        summary_first_dropout = self.pytorch_params['transformer_summary_first_dropout'] if 'transformer_summary_first_dropout' in self.pytorch_params.keys() else 0.1
        attention_dropout = self.pytorch_params['transformer_attention_dropout'] if 'transformer_attention_dropout' in self.pytorch_params.keys() else 0.1
        dropout = self.pytorch_params['transformer_dropout'] if 'transformer_dropout' in self.pytorch_params.keys() else 0.2
        self.logger.info(f"Transformer's summary first dropout : {summary_first_dropout}")
        self.logger.info(f"Transformer's attention dropout : {attention_dropout}")
        self.logger.info(f"Transformer's dropout : {dropout}")

        # Update pytorch_params for saving purposes
        self.pytorch_params['transformer_summary_first_dropout'] = summary_first_dropout
        self.pytorch_params['transformer_attention_dropout'] = attention_dropout
        self.pytorch_params['transformer_dropout'] = dropout

        # Get transformer path
        transformers_path = utils.get_transformers_path()
        transformer_path = os.path.join(transformers_path, self.transformer_name)

        # Check if available locally
        if os.path.exists(transformer_path):
            self.logger.info("Use the local transformer.")
            final_transformer_localisation = transformer_path
        else:
            self.logger.warning("Can't find the local transformer.")
            self.logger.warning("We try to get it from the web.")
            final_transformer_localisation = self.transformer_name

        # AutoConfig base on labels
        config = AutoConfig.from_pretrained(
            pretrained_model_name_or_path=final_transformer_localisation,
            summary_first_dropout=summary_first_dropout,
            attention_dropout=attention_dropout,
            dropout=dropout,
            num_labels=len(self.list_classes),
            id2label=self.dict_classes,
            label2id={v: i for i, v in self.dict_classes.items()},
        )

        # Retrieve transformer model
        transformer_model = AutoModelForSequenceClassification.from_pretrained(final_transformer_localisation, config=config)

        # Resize embedding if some special tokens added
        if self.tokenizer_special_tokens:
            transformer_model.resize_token_embeddings(len(self.tokenizer))

        # Allow full retraining
        for param in transformer_model.parameters():
            param.requires_grad = True

        # Returns tranformer model
        return transformer_model

    def _get_model(self, train_dataloader_size: Union[int, None] = None) -> Any:
        '''Gets a model structure

        Kwargs:
            train_dataloader_size (int): number of batch per epochs. Useful to set a learning rate scheduler
        Returns:
            (?): a PyTorch model
        '''
        if self.tokenizer is None:
            raise AttributeError("The tokenizer is not set. Can't get the model.")
        # Retrieve transformer
        transformer_model = self._get_transformer()

        # Set parameters
        lr = self.pytorch_params.get('learning_rate', 2.0e-05)
        decay = self.pytorch_params.get('decay', 0.0)
        adam_epsilon = self.pytorch_params.get('adam_epsilon', 1.5e-06)
        gradient_clip_val = self.pytorch_params.get('gradient_clip_val', 1.0)
        warmup_proportion = self.pytorch_params.get('warmup_proportion', 0.2)
        lr = self.pytorch_params.get('learning_rate', 2.0e-05)
        decay = self.pytorch_params.get('decay', 0.0)
        adam_epsilon = self.pytorch_params.get('adam_epsilon', 1.5e-06)
        gradient_clip_val = self.pytorch_params.get('gradient_clip_val', 1.0)
        warmup_proportion = self.pytorch_params.get('warmup_proportion', 0.2)
        run_lr_scheduler = self.pytorch_params.get('run_lr_scheduler', False)
        self.logger.info(f"Learning rate utilisée : {lr}")
        self.logger.info(f"Decay utilisé : {decay}")
        self.logger.info(f"Adam's epsilon utilisé : {adam_epsilon}")
        self.logger.info(f"Gradient clipping utilisé : {gradient_clip_val}")
        if run_lr_scheduler:
            self.logger.info("Learning rate scheduler used")
            self.logger.info(f"Warmup proportion used : {warmup_proportion}")

        # Update pytorch_params for saving purposes
        self.pytorch_params['learning_rate'] = lr
        self.pytorch_params['decay'] = decay
        self.pytorch_params['adam_epsilon'] = adam_epsilon
        self.pytorch_params['gradient_clip_val'] = gradient_clip_val
        self.pytorch_params['warmup_proportion'] = warmup_proportion
        self.pytorch_params['run_lr_scheduler'] = run_lr_scheduler

        # Loss function
        loss_func = utils_deep_torch.focal_loss if self.multi_label else utils_deep_torch.categorical_crossentropy_from_logits
        metrics_func = self.get_metrics_simple_multilabel if self.multi_label else self.get_metrics_simple_monolabel

        # Return
        model = TaskClass(multi_label=self.multi_label, transformer_model=transformer_model,
                          tokenizer=self.tokenizer, loss_func=loss_func, probas_to_classes=self.get_classes_from_proba,
                          get_metrics=metrics_func, output_path=self.model_dir,
                          epochs=self.epochs, lr=lr, decay=decay, adam_epsilon=adam_epsilon,
                          gradient_clip_val=gradient_clip_val, run_lr_scheduler=run_lr_scheduler,
                          warmup_proportion=warmup_proportion, train_dataloader_size=train_dataloader_size)
        return model

    def _convert_inputs(self, x, y=None) -> Tuple[Any, Any, Any]:
        '''Converts (x, y) data to the model's input format

        Args:
            x (?): Array-like, shape = [n_samples, n_features]
        Kwargs:
            y (?): Array-like, shape = [n_samples, n_features]
        '''
        # First, cast to numpy arrays
        x = np.array(x)
        if y is not None:
            y = np.array(y)

        # Get features from tokenizer
        features = []
        for i in tqdm(range(len(x))):
            text = x[i]
            label = y[i] if y is not None else None
            entry = self.tokenizer(text, padding=self.padding, truncation=self.truncation, max_length=self.max_sequence_length)
            features.append((entry["input_ids"], entry["attention_mask"], label))

        # Retrieve inputs, masks & labels
        all_input_ids = torch.tensor([f[0] for f in features], dtype=torch.long)
        all_input_mask = torch.tensor([f[1] for f in features], dtype=torch.long)
        if y is not None:
            all_label = torch.tensor([f[2] for f in features], dtype=torch.float32)
        else:
            all_label = None
        return all_input_ids, all_input_mask, all_label

    def _get_train_dataloader(self, batch_size: int, x_train, y_train_dummies = None) -> DataLoader:
        '''Prepares the input data for the model

        Args:
            batch_size (int): Train batch size
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train_dummies (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (DataLoader): Data loader
        '''
        all_input_ids, all_input_mask, all_label = self._convert_inputs(x_train, y_train_dummies)
        if all_label is not None:
            train_dataset = TensorDataset(all_input_ids, all_input_mask, all_label)
        else:
            raise ValueError("Aucun label associé au training set...")
        train_dl = DataLoader(train_dataset, num_workers=0, batch_size=batch_size, sampler=RandomSampler(train_dataset))
        return train_dl

    def _get_test_dataloader(self, batch_size: int, x_test, y_test_dummies=None) -> DataLoader:
        '''Prepares the input data for the model

        Args:
            x_test (?): Array-like, shape = [n_samples, n_features]
            batch_size (int): Test batch size
        Kwargs:
            y_test_dummies (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (DataLoader): Data loader
        '''
        all_input_ids, all_input_mask, all_label = self._convert_inputs(x_test, y_test_dummies)
        if all_label is not None:
            test_dataset = TensorDataset(all_input_ids, all_input_mask, all_label)
        else:
            test_dataset = TensorDataset(all_input_ids, all_input_mask)
        test_dl = DataLoader(test_dataset, num_workers=0, batch_size=batch_size)
        return test_dl

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save configuration JSON
        if json_data is None:
            json_data = {}

        # Add specific data
        json_data['transformer_name'] = self.transformer_name
        json_data['max_sequence_length'] = self.max_sequence_length
        json_data['tokenizer_special_tokens'] = self.tokenizer_special_tokens
        json_data['padding'] = self.padding
        json_data['truncation'] = self.truncation

        # Save
        super().save(json_data=json_data)

    def reload_model(self, model_path: str, **kwargs) -> Any:
        '''Reloads a model saved with ModelCheckpoint

        Args:
            model_path (str): Checkpoint's full path
        Kwargs:
            kwargs: Dict of kwargs to override predefined params (TO BE CHECKED !!!)
        Returns:
            ?: Pytorch lightning model
        '''
        model = TaskClass.load_from_checkpoint(model_path, **kwargs)
        # Idea to test in order not to have mandatory hyperparameters savings
        # 1. Only save the weights of the model
        # 2. Save the tokenizer
        # 3. Reload :
        #   - Reload the weights (torch.load ...)
        #   - Reload the tokenizer
        #   - Initialize the class with these arguments

        # Set trained to true if not already true
        if not self.trained:
            self.trained = True
            self.nb_fit = 1

        return model

    def reload_from_standalone(self, **kwargs) -> None:
        '''Reloads a model from its configuration and "standalones" files
        - /!\\ Experimental /!\\ -

        Kwargs:
            configuration_path (str): Path to configuration file
            checkpoint_path (str): Checkpoint's full path
        Raises:
            ValueError: If configuration_path is None
            ValueError: If checkpoint_path is None
            FileNotFoundError: If the object configuration_path is not an existing file
            FileNotFoundError: If the object checkpoint_path is not an existing file
        '''
        # Retrieve args
        configuration_path = kwargs.get('configuration_path', None)
        checkpoint_path = kwargs.get('checkpoint_path', None)

        if configuration_path is None:
            raise ValueError("The argument configuration_path can't be None")
        if checkpoint_path is None:
            raise ValueError("The argument checkpoint_path can't be None")
        if not os.path.exists(configuration_path):
            raise FileNotFoundError(f"The file {configuration_path} does not exist")
        if not os.path.exists(checkpoint_path):
            raise FileNotFoundError(f"The file {checkpoint_path} does not exist")

        # Load confs
        with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        # Can't set int as keys in json, so need to cast it after reloading
        # dict_classes keys are always ints
        if 'dict_classes' in configs.keys():
            configs['dict_classes'] = {int(k): v for k, v in configs['dict_classes'].items()}
        elif 'list_classes' in configs.keys():
            configs['dict_classes'] = {i: col for i, col in enumerate(configs['list_classes'])}

        # Set class vars
        # self.model_name = # Keep the created name
        # self.model_dir = # Keep the created folder
        self.nb_fit = configs.get('nb_fit', 1)  # Consider one unique fit by default
        self.trained = configs.get('trained', True)  # Consider trained by default
        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['x_col', 'y_col',
                          'list_classes', 'dict_classes', 'multi_label', 'level_save',
                          'batch_size', 'epochs', 'validation_split', 'patience',
                          'embedding_name', 'max_sequence_length', 'tokenizer_special_tokens',
                          'padding', 'truncation', 'pytorch_params', 'transformer_name']:
            setattr(self, attribute, configs.get(attribute, getattr(self, attribute)))

        # Reload model
        self.model = self.reload_model(checkpoint_path)
        # Do not forget to freeze
        self.model.freeze()  # type: ignore

        # Save best model in new folder
        new_checkpoint_path = os.path.join(self.model_dir, 'best_model.ckpt')
        shutil.copyfile(checkpoint_path, new_checkpoint_path)

        # Reload tokenizer
        self.tokenizer = self.model.tokenizer  # type: ignore


# Set task (i.e. our model)
class TaskClass(pl.LightningModule):
    @no_type_check  # Mypy does not detect changes in hparams
    def __init__(self, multi_label: bool, transformer_model, tokenizer, loss_func: Callable, probas_to_classes: Callable,
                 get_metrics: Callable, output_path: str, epochs: int = 5, lr: float = 2.0e-05, decay: float = 0.0,
                 adam_epsilon: float = 1.5e-06, gradient_clip_val: Union[float, int] = 1.0,
                 run_lr_scheduler: bool = False, warmup_proportion: float = 0.2,
                 train_dataloader_size: Union[int, None] = None) -> None:
        '''Initialization task class

        Args:
            multi_label: If working on a multilabel problem
            transformer_model: Transformer model loaded
            tokenizer: Tokenizer associated with the transformer model
            loss_func (callable): Loss function to be used
            probas_to_classes (callable): Function to get classes from probas
            get_metrics (callable): Function to get metrics
            output_path (str): Model directory
        Kwargs:
            epochs (int): Number of epochs to use
            lr (int): Initial learning rate to use
            decay (float): Decay to use
            adam_epsilon (float): Adam's epsilon to use
            gradient_clip_val (int | float): Gradient clipping value to use
            run_lr_scheduler (bool): Whether a learning rate scheduler should be used
            warmup_proportion (float): Warmup proportion to use (for learning rate scheduler)
            train_dataloader_size (int): Number of batch per epochs. Useful to set a learning rate scheduler
        '''
        super(TaskClass, self).__init__()
        self.save_hyperparameters("multi_label", "transformer_model", "tokenizer", "loss_func",
                                  "probas_to_classes", "get_metrics", "output_path",
                                  "epochs", "lr", "decay", "adam_epsilon", "gradient_clip_val",
                                  "run_lr_scheduler", "warmup_proportion", "train_dataloader_size")
        self.multi_label: bool = self.hparams.multi_label
        self.model: Any = self.hparams.transformer_model
        self.tokenizer: Any = self.hparams.tokenizer
        self.loss_func: Callable = self.hparams.loss_func
        self.probas_to_classes: Callable = self.hparams.probas_to_classes
        self.get_metrics: Callable = self.hparams.get_metrics
        self.output_path: str = self.hparams.output_path
        self.epochs: int = self.hparams.epochs
        self.lr: float = self.hparams.lr
        self.decay: float = self.hparams.decay
        self.adam_epsilon: float = self.hparams.adam_epsilon
        self.gradient_clip_val: Union[int, float] = self.hparams.gradient_clip_val
        self.run_lr_scheduler: bool = self.hparams.run_lr_scheduler
        self.warmup_proportion: float = self.hparams.warmup_proportion
        self.train_dataloader_size: int = self.hparams.train_dataloader_size

    def to_device(self, obj) -> Any:  # Useful ?
        return obj.cuda() if torch.cuda.is_available() else obj

    def convert_network_to_device(self) -> None:
        if torch.cuda.is_available():
            self.to('cuda')
        else:
            self.to('cpu')

    def configure_optimizers(self) -> dict:
        '''Setup optimizers'''
        param_optimizer = list(self.model.named_parameters())
        no_decay = ["bias", "gamma", "beta"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
                "weight_decay_rate": self.decay,
            },
            {
                "params": [p for n, p in param_optimizer if any(nd in n for nd in no_decay)],
                "weight_decay_rate": 0.0,
            },
        ]
        optimizer = AdamW(optimizer_grouped_parameters, lr=self.lr, eps=self.adam_epsilon)
        return_dict = {'optimizer': optimizer}

        # Add LR scheduler if needed
        if self.run_lr_scheduler and self.train_dataloader_size is not None:
            total_steps = int(self.train_dataloader_size * self.epochs)
            warmup_steps = self.warmup_proportion
            if warmup_steps <= 1:
                warmup_steps = int(total_steps * warmup_steps)
            else:
                warmup_steps = int(warmup_steps)  # Typing
            # TODO : to be improved -> allow other LR schedulers
            lr_scheduler = LambdaLR(optimizer, lr_lambda=utils_deep_torch.LRScheduleWithWarmup(warmup_steps=warmup_steps, total_steps=total_steps))
            return_dict['lr_scheduler'] = {}
            return_dict['lr_scheduler']['scheduler'] = lr_scheduler
            return_dict['lr_scheduler']['interval'] = 'step' # Update at each step
            return_dict['lr_scheduler']['frequency'] = 1
        # Return dict
        return return_dict

    def forward(self, input_ids, attention_mask=None):
        '''Forward step'''
        x = self.model(input_ids=input_ids, token_type_ids=None, attention_mask=attention_mask, head_mask=None)
        return x[0]  # Return logits

    def training_step(self, batch, batch_idx):
        '''Training step'''
        input_ids, attention_mask, label = batch
        logits = self.forward(input_ids, attention_mask)
        loss = self.loss_func(logits, label)
        lr = self.trainer.optimizers[0].param_groups[0]["lr"]
        self.log("lr", lr, prog_bar=True, on_step=True, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        '''Validation step'''
        input_ids, attention_mask, label = batch
        logits = self.forward(input_ids, attention_mask)
        loss = self.loss_func(logits, label)
        outputs = {
            "val_loss": loss,
            "logit": logits.cpu().numpy(), # Useful for metrics
            "label": label.cpu().numpy(), # Useful for metrics
        }
        return outputs

    def validation_epoch_end(self, outputs) -> None:
        '''On epochs ends, gets validation metrics'''
        avg_loss = torch.stack([x["val_loss"] for x in outputs]).mean() # val_loss = loss mean among batches
        y_true = np.vstack([x["label"] for x in outputs])
        logits = np.vstack([x["logit"] for x in outputs])
        probas = self.get_probas_from_logits(logits)
        y_pred = np.array(list(map(self.probas_to_classes, probas)))
        # If not multi-labels, retrieves classes from OHE vectors
        if not self.multi_label:
            y_true = np.array(list(map(self.probas_to_classes, y_true)))
        # Get metrics
        metrics_df = self.get_metrics(y_true, y_pred)
        all_metrics = metrics_df[metrics_df.Label == 'All']
        # Logs metrics
        self.log("val_loss", avg_loss, prog_bar=True, on_step=False, logger=True)
        for col in all_metrics.columns:
            if col != 'Label' and all_metrics[col].values[0] != None:
                self.log(f"val_{col}", all_metrics[col].values[0], prog_bar=True, on_step=False, logger=True)

    def test_step(self, batch, batch_idx):
        '''Test step'''
        input_ids, attention_mask = batch[:2]
        logits = self.forward(input_ids, attention_mask)
        outputs = {"logit": logits.cpu().numpy()}
        return outputs

    def test_epoch_end(self, outputs) -> None:
        '''On epochs ends, gets test metrics'''
        logits = np.vstack([x["logit"] for x in outputs])
        probas = self.get_probas_from_logits(logits)
        setattr(self, "test_probas", probas)
        # Nothing to log ?

    def get_probas_from_logits(self, logits) -> np.ndarray:
        '''Function to retrieve probas from logits (in numpy format)'''
        if self.multi_label:
            probas = Sigmoid()(torch.from_numpy(logits)).numpy()
        else:
            probas = Softmax(dim=-1)(torch.from_numpy(logits)).numpy()
        return probas


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")
